# Script to take formatted AGES input files and convert them into a nice shiny table
# Based on original "formatta.c" by Luca Cortese
# Input files should consist of the input parameters to mbspect (one per line), the output
# from miriad to the terminal (i.e. the measured parameters of the source) and the spectra
# from the miriad log file
# This version checks for the strange \r carriage returns, which Python sometimes insists are
# present even though they're bloody well not.
# Produces standard catalogue table of the format :
# Hid  Name  HH:MM:SS  R.A.(deg)  errRA(SS)  DD:MM:SS  Dec.(deg)  errDec(SS)  velocity  errvel  W50  errW50  W20  errW20  PeakFlux  errPeak  TotFlux  errTot  MHI  SNpeak  SNtot  rms  Field     
# Also produces two html files so that the user can query NED ans the SDSS for each source.

# Key :
# Hid  : standard unpronounceable, useless telephone number of format JHHMMSS+DDMMSS
# Name : human-readable ID, based on the file name. Assumed to be of the format NAME_XXX_X.txt

# Coordinates : these are found by checking both the input parmeter line "coord=..." and the
# position-fitting part of the spectrum P4 and P5. Checks also if "options=posfit...". If it does
# not, then only input coordinates are used (so you don't need to manually insert this section if
# position fitting failed). Also checks if position fitting was used but failed, both by looking
# for the standard miriad error mesaage and by directly comparing the user's input coordinates with
# the position-fitted coordinates (sometimes they can strongly mismatch without miriad giving any
# error !). If they disagree by more than 3.5', it revers to the user's parameters and prints a
# warning to the terminal.

# Coordinate errors : if position fitting is not used (or failed), a blank value is written to the 
# column. Otherwise uses the P3 "positional errors" values with the following limitations as used 
# by Luca :
# errRA= sqrt(errRA*errRA + 10*10)/15 (this gives error in time, not degrees)
# errDec= sqrt(errDec*errDec + 10*10)

# Velocity : uses (following Luca, I think) the maximised W50 line center (MXC) if it exists.
# If it does not, uses the position of maximum flux in the profile range, MX. Should be very robust
# to measurement failures if profile range has to be limited.

# Velocity error : following Luca, uses errVel = (3*errPeak/Fpeak)*sqrt(10*0.5*(W20-W50)
# The component parameters are extracted as described below.

# W50 : uses the maximised value, MXW
# W50 error : follow Luca, this is 2*errVel (WHY ?)
# W20 values follow the same procedure, but error is 3*errVel

# Peak flux : following Luca this is not extracted directly but by SNpeak * rms, units of mJy
# Peak flux error : errPeak = sqrt(rms*rms+ 0.05*0.05*Fpeak*Fpeak), units of mJy
 
# Total flux : this comes from "Robust moment 0", i.e. "MR" line, units of Jy
# Total flux error : errTot = (4*errPeak/Fpeak)*sqrt(10*Tflux*Fpeak), units of Jy

# MHI = log10(2.36E5*(H/71)^2 * Tflux) in logarithmic solar masses

# S/N peak : directly extracted from SN1

# S/N total : directly extracted from SN2

# rms : directly extracted from "FR clipped rms", in mJy

# Field : manually specified


import math
from math import *
import re

# *** USER-ADJUSTABLE PARAMETERS ***
# User must create a list of the files they wish to tabulate. Files should have a consistent naming
# convention, NAME_XXX_logspec.txt
filelist = 'FileList'

# User also needs to specify the name of the field
fieldname = 'NGC7619'

# User can also specify the name of the output table if they like; default is to use the field name
outfile1 = open(fieldname+'_HItable.txt','w')

# Also creates handy SDSS and NED html pages for easy queries
outfile2 = open(fieldname+'_SDSS.html','w')
outfile3 = open(fieldname+'_NED.html','w')

# Finally, user should set the value for H0 in km/s/Mpc
H = 71.0
# *** END OF USER-ADJUSTABLE PARAMETERS ***




# Write the header to the table file
header = '#Hid           Name      RA(HHMMSS)  R.A.(deg)  eR(SS) Dec(DDMMSS) Dec.(deg)  eD(SS) Vel  eV   W50 eW50 W20 eW20 PkFlux ePkF TotFlux eTotF MHI   SNpeak SNtot rms Field'+'\n'     
outfile1.write(header)

speclist = open(filelist,'r')

for line in speclist:
	infilename = line.rstrip('\n')	# Remove carriage return
	print 'Processing ',infilename
		
	# Set a flag for position fitting
	posfit = False
	
	# Assume that no other variables exist at all. Setting them here prevents them from being
	# carried over when the previous file contained them but the current one doesn't.
	RAC      = None
	DecC     = None
	HHu      = None
	MMu      = None
	SSu      = None
	DDu      = None
	MMdu     = None
	SSdu     = None
	errRA    = None
	errDec   = None
	RApfit   = None
	HHp		= None
	MMp		= None
	SSp		= None
	Decpfit	= None
	DDp		= None
	MMdp		= None
	SSdp		= None
	velocity = None
	W50		= None
	W20		= None
	Tflux 	= None
	peakSN	= None
	totalSN	= None
	rms 		= None
	
	# Go through each input file, getting the relevant parameters and writing them to the files
	infile = open(infilename,'r')
	for fline in infile:
		fline = fline.rstrip('\n')
		# Get user-input coordinates. These may be over-written later if position fitting used and
		# deviation between the two is not excessive.
		if ('coord' in fline) and ('=' in fline):
			RAC  = fline.split(',')[0].split('=')[1].rstrip('\n')
			DecC = fline.split(',')[1]
			
			# Sexigesimal columns for database table.
			# HH may have trailing spaces, so remove by first converting it to an integer 
			HHu = str(int(RAC.split(':')[0]))
			MMu = str(RAC.split(':')[1])
			SSu = str(RAC.split(':')[2]).rstrip('\n')   # Remove carriage return			
			 
			DDu = str(int(DecC.split(':')[0]))
			MMdu = str(DecC.split(':')[1])
			SSdu = str(DecC.split(':')[2]).rstrip('\n') # Remove carriage return
			
		if 'options' in fline:
			if 'posfit' in fline:
				posfit = True
		
		# Get position-fitting errors, if they exist (set manually to blank later if they don't)
		if 'P3' in fline:
			# Split first by colon, then by any white space
			# Remember split generates a list, so have to specify which component to split when doing
			# multiple splits
			errRA  = fline.split(':')[1].split()[0]
			errDec = fline.split(':')[1].split()[1]
			
			errRA = float(errRA)
			errDec = float(errDec)
						
			errRA  = sqrt(errRA*errRA + 10.0*10.0)/15.0
			errDec = sqrt(errDec*errDec + 10.0*10.0)

			
			# If errors are very large (> 3.5'), print a warning to the screen
			# Set values to be blank strings "" (plus a space for table formating purposes)
			if (errDec >= (3.5*60.0)) or ((errRA*15.0) >= (3.5*60)):
				print 'WARNING ! Position errors are very large, check these carefully !'
				print 'Over-riding position errors in input file and setting blank values instead.'
				#raw_input('Press any key to continue...')
									# Alternatively use default minimums
				errRA = '"" '	# 0.7
				errDec ='"" '	# 10.0
				
			if 'nan' in str(errRA):
				print 'WARNING ! RA position error not found.'
				errRA = '"" '	#0.7
				
			if 'nan' in str(errDec):
				print 'WARNING ! Dec position error not found.'
				errDec = '"" '	#10.0
			
			# Alternative approach might be to produce two tables, one for the database upload and another
			# for actual use - but it's probably easy enough to edit the main table instead.
									
		
		# If position fitting used, get position-fit coordinates. These will be used in the table
		# unless they deviate strongly from the user inputs, in which case an error will be raised.
		if (posfit == True) and ('P4' in fline):
			RApfit = fline.split('ion:')[1]
			HHp = str(int(RApfit.split(':')[0]))
			MMp = str(RApfit.split(':')[1])
			SSp = str(RApfit.split(':')[2]).rstrip('\n')
						
		if (posfit == True) and('P5' in fline):
			Decpfit = fline.split('ion:')[1]
			DDp = str(int(Decpfit.split(':')[0]))
			MMdp = str(Decpfit.split(':')[1])
			SSdp = str(Decpfit.split(':')[2]).rstrip('\n')
			
		
		# Final position parameter check - if 'Position error seems large' was written, that means position fitting
		# failed. Set posfit to be false - next section will use this to set coordinates and errors from the user's
		# parameters, which have already been recorded.
		if 'Position error seems large' in fline:
			print 'Posfit failed ! Reverting to user-input coordinates.'
			posfit = False
		
		# Get the velocity. Will always be able to do this by peak flux, but usually will get from
		# line centre.
		if ('MX' in fline) and ('Maximum' in fline):
			velocity = float(fline.split()[5])
		if 'MXC' in fline:
			velocity = float(fline.split()[3])
		
		# Get the W50 and W20
		if 'MXW' in fline:
			W50 =	float(fline.split()[3])
			W20 = float(fline.split()[4])
			
		# Get total flux
		if 'MR' in fline:
			Tflux = float(fline.split()[1])
		
		# Get peak S/N
		if 'SN1' in fline:
			peakSN = float(fline.split()[5])
		
		# Get total S/N
		if 'SN2' in fline:
			totalSN = fline.split()[5]
	
		# Get rms
		if 'FR' in fline:
			rms = float(fline.split()[3])

			
	
	# Once have read all the necessary values from the file, exit the loop and prepare to tabulate
	# the data
	
	decRAu = (float(HHu)+(float(MMu)/60.0)+(float(SSu)/3600.0))*15.0	# User RA
	decDecu = float(DDu)+(float(MMdu)/60.0)+(float(SSdu)/3600.0)	# User Dec
			
	# If position fitting is false, set values for position errors to be blank strings
	if posfit == False:
		print 'Position fitting NOT used !'
		errRA = '"" '	# Single quotes enclose double quotes to force them to be a string
		errDec = '"" ' # Space added for nicer table formatting
			
		# But for the database table we will use the minimum values
		#errRAdb = 10.0/15.0
		#errDecdb = 10.0
			
	# If position fitting was used, compare fitted position and user input. If deviation > 3.5', 
	# raise an error.
	if posfit == True:		
		decRAp = (float(HHp)+(float(MMp)/60.0)+(float(SSp)/3600.0))*15.0	# Posfit RA
		decDecp = float(DDp)+(float(MMdp)/60.0)+(float(SSdp)/3600.0)	# Posfit Dec
		
		# Get angular separation in degrees
		angsep = sqrt((decRAu-decRAp)**2.0 + (decDecu-decDecp)**2.0)
				
		# If there's a problem with the position fitting not matching the user's input, raise an 
		# error and stop.
		if angsep*60.0 >= 3.5:
			print 'ERORR ! User input and position fit coordinates do not match !'
			print 'Failed on file ',infilename
			print 'Angular separation =',angsep*60.0,'arcminutes'
			exit()
			
	# As long as the user coordinates and positing-fitted values agree reasonably well, proceed
	# Set final values to output to table as position-fitted values if position fitting was done
	if posfit == True:
		decimalRA = decRAp
		decimalDec= decDecp
		HHr = HHp
		MMr = MMp
		SSr = SSp
		sexRA = HHr+':'+MMr+':'+SSr
		
		DDd = DDp
		MMd = MMdp
		SSd = SSdp
		sexDec = DDd+':'+MMd+':'+SSd
			
	# If position fitting false, use user-specified values
	if posfit == False:
		decimalRA = decRAu
		decimalDec= decDecu
		HHr = HHu
		MMr = MMu
		SSr = SSu
		sexRA = HHr+':'+MMr+':'+SSr
			
		DDd = DDu
		MMd = MMdu
		SSd = SSdu
		sexDec = DDd+':'+MMd+':'+SSd
		
			
	# Add trailing zeros to coordinates
	HHr = HHr.zfill(2)
	MMr = MMr.zfill(2)
	
	DDd = DDd.zfill(2)
	MMd = MMd.zfill(2)
	
	# Sometimes we'll be using values from the position-fitted coordinates and sometimes from the user inputs.
	# Position-fitted coordinates ALWAYS have the format HH:MM:SS.ss. But user values do not : the seconds value
	# can simply be S. So 4 becomes 00004. Could also have 4.4, which would become 004.4, or 4.40 which would 
	# correctly become 04.40, or 04.4, which would incorrectly become 004.4.
	# Robust solution : split the string by the '.'. If there is one, take the integer parts. If they are < 10,
	# then convert back to string with zfill(2). If there is not a '.', take the integer part, add the trailing
	# zeros and then add another string '.00'.
	# This only needs to be done for the seconds, as simple zfill will be sufficient on the hours and minutes.
	# This is only a formatting issue for the table - the values for the decimal format positions should be 
	# correct.
	if '.' in SSr:
		SSr1 = SSr.split('.')[0].zfill(2)
		SSr2 = SSr.split('.')[1].zfill(2)
		SSr = SSr1+'.'+SSr2
	if '.' not in SSr:
		SSr = SSr.zfill(2)+'.00'
		
	if '.' in SSd:
		SSd1 = SSd.split('.')[0].zfill(2)
		SSd2 = SSd.split('.')[1].zfill(2)
		SSd = SSd1+'.'+SSd2
	if '.' not in SSd:
		SSd = SSd.zfill(2)+'.00'	
		
	#Peak flux
	Pflux = peakSN*rms
	
	# Error on peak flux
	errPeak = sqrt(rms*rms+ 0.05*0.05*Pflux*Pflux)
	
	# Error on total flux	
	errTot = (4.0*errPeak/Pflux)*sqrt(10.0*Tflux*Pflux)
	
	# Error on velocity. Widths may be undefined - miriad sets these to zero.
	if (W50 > 0.0) and (W20 > 0.0):
		errVel = (3.0*errPeak/Pflux)*sqrt(10.0*0.5*(W20-W50))
	if (W50 == None) or (W50 == 0.0) or (W20 == 0.0) or (W20 == None):
		errVel = None
		
	# Error on velocity widths
	if errVel > 0.0:
		errW50 = 2.0*errVel
		errW20 = 3.0*errVel
	
	# Have to force width errors to be None if velocity error is None, otherwise it will use the previous values	
	if errVel == None:
		errW50 = None
		errW20 = None
	
	# HI mass	
	MHI = log10(2.36E5*(velocity/H)**2.0 * Tflux)
	
	# Telephone number ID
	Hid = 'J'+str(HHr)+str(MMr)+str(SSr.split('.')[0])+'+'+str(DDd)+str(MMd)+str(SSd.split('.')[0])
	
	# Human readable ID
	Name = infilename.split('_')[0]+'_'+infilename.split('_')[1]
	
	# Finally, any parameter that could still be "None" should be replaced with "" for final output.
	# Most parameters can actually be determined one way or another : peak flux, velocity, position, and 
	# associated errors are all well-defined. The exceptions are velocity widths and, consequently, the errors on
	# the systemic velocity and velocity widths.
 	if (W50 == None) or (W20 == None):
		errVel = '""'
		errVel = errVel+' '*(3-len(errVel))
		
		errW50 = '""'
		errW50 = errW50+' '*(4-len(errW50))
		
		errW20 = '""'
		errW20 = errW20+' '*(4-len(errW20))	
	
	
	# Now we can write everything to the tables ! But make sure to format number lengths properly
	# and to write blank values if variables are set to "None".
	# REFERENCE NOTE : "%.5f" % thing will give 'thing' to 5 decimal places
	# Multiplying a string will duplicate that string and join them together, so 5*'x' would be xxxxx and 5*' '
	# gives 5 spaces.
	decimalRA = str(decimalRA)
	decimalRA = decimalRA+' '*(13-len(decimalRA)) 	# Add extra spaces to keep the length in the table the same
	decimalDec= str(decimalDec)
	decimalDec = decimalDec+' '*(13-len(decimalDec)) # Add extra spaces to keep the length in the table the same
	# If the errors are floats, convert them to 1 d.p. strings. If they're not floats they're probably the blank
	# strings, "", so can be left alone.
	if type(errRA) is float:
		errRA = str("%.1f" % errRA) 
	decimalDec = str(decimalDec)
	if type(errDec) is float:
		errDec = str("%.0f" % errDec)
	errDec = errDec+' '*(3-len(str(errDec)))	# Extra spaces for table
	velocity = str("%.0f" % velocity)
	velocity = velocity+' '*(5-len(velocity))
	
	# Several different possibilities for velocity and width errors. If the velocity error is None, the width
	# errors will also be none. This occurs when either W50 or W20 is 0. But if they are both measurable then
	# everything should be fine, since W20 is always > W50.
	if (W50 == 0.0) or (W20 == 0.0):
		errVel = '"" '
		errVel = errVel+' '*(3-len(errVel))
		
		errW50 = '""'
		errW50 = errW50+' '*(4-len(errW50))
		
		errW20 = '""'
		errW20 = errW20+' '*(4-len(errW20))
		
		# And also set the widths themselves - cannot be zero, should be blank
		if W50 == 0.0:
			W50 = '""'
			W50 = W50+' '*(3-len(W50))
			
		if W20 == 0.0:
			W20 = '""'
			W20 = W20+' '*(3-len(W20))
	
	# If both widths measured then errors will automatically be valid as well, just need to format correctly	
	# But, Python evaluates "" to be greater than zero, so have to check for the undefined case.
	if (W50 > 0.0) and (W20) > 0.0 and ('"' not in str(W50)) and ('"' not in str(W20)):
		errVel = str("%.0f" % errVel)
		errVel = errVel+' '*(3-len(errVel))
		
		errW50 = str("%.0f" % errW50)
		errW50 = errW50+' '*(4-len(errW50))
		
		errW20 = str("%.0f" % errW20)
		errW20 = errW20+' '*(4-len(errW20))

		W50 = str("%.0f" % W50)
		W50 = W50+' '*(3-len(W50))
		
		W20 = str("%.0f" % W20)
		W20 = W20+' '*(3-len(W20))

	# Final checks - above cases can miss if only one of W50 or W20 is "" or None
	if type(W50) is not str:
		if W50 is None:
			W50 = '""'
			W50 = W50+' '*(3-len(W50))
		else:
			W50 = str("%.0f" % W50)
			W50 = W50+' '*(3-len(W50))
	if type(W20) is not str:
		if W20 is None:
			W20 = '""'
			W20 = W20+' '*(3-len(W20))
		else:
			W20 = str("%.0f" % W20)
			W20 = W20+' '*(3-len(W20))

	# Peak flux in mJy
	Pflux = str("%.1f" % (Pflux/1E-3))
	Pflux = Pflux+' '*(6-len(Pflux))
	
	errPeak = str("%.1f" % (errPeak/1E-3))
	errPeak = errPeak+' '*(4-len(errPeak))
	
	Tflux = str(Tflux)
	Tflux = Tflux+' '*(7-len(Tflux))
	
	errTot = str("%.3f" % errTot)
	errTot = errTot+' '*(4-len(errTot))
	
	MHI = str("%.2f" % MHI)
	MHI = MHI+' '*(5-len(MHI))
	
	peakSN = str("%.1f" % peakSN)
	peakSN = peakSN+' '*(6-len(peakSN))
	
	totalSN = str("%.1f" % float(totalSN))
	totalSN = totalSN+' '*(5-len(totalSN))
	
	rms = str("%.1f" % (rms/1E-3))
			
	string1 = str(Hid)+' '+str(Name)+' '+HHr+':'+MMr+':'+SSr+' '+decimalRA+' '+errRA+' '+DDd+':'+MMd+':'+SSd+' '+decimalDec+' '+errDec+' '+velocity+' '+errVel+' '+W50+' '+errW50+' '+W20+' '+errW20+' '+Pflux+' '+errPeak+' '+Tflux+' '+errTot+' '+MHI+' '+peakSN+' '+totalSN+' '+rms+' '+fieldname+'\n'
	string2 = '<a href="http://skyserver.sdss.org/dr13/en/tools/chart/image.aspx?ra='+decimalRA+'&dec='+decimalDec+'&scale=0.3961&width=515&height=515">'+Name+' </a>  <br/>'+'\n'
	string3 = '<a href="http://ned.ipac.caltech.edu/cgi-bin/objsearch?in_csys=Equatorial&inequinox=J2000.0&lon='+str(float(decimalRA)/15.0)+'&lat='+decimalDec+'&radius=1.75&search_type=Near+Position+Search">'+Name+'  </a>  <br/>'+'\n'
		
	# For the removal of these strange, mysterious \r carriage returns. I have absolutely no idea
	# where these come from, but they're a bloody nusiance. 
	newstring1 = ''
	for a in string1:
		if '\r' not in a:
			newstring1 = newstring1+a
	#print 'done !'
	
	outfile1.write(newstring1)
	outfile2.write(string2)
	outfile3.write(string3)

	# CHECK IF WILL RUN WITHOUT POSITION FITTING SECTION; ANY CONFLICTS IF SECTION IS PRESENT BUT POSFIT NOT
	# ACTUALLY USED.
			
	infile.close()	
	
		
print 'Done ! Check the above output for warnings.'		
	
speclist.close()
outfile1.close()
outfile2.close()

